import json
from unittest.mock import ANY

import pytest

from duplo_resource.service import DuploService
from duplocloud.errors import DuploError


invalid_kwargs = [
  {
    'name': 'widget',
  },
  {
    'name': 'widget',
    'image': 'nginx:latest',
    'container_image': [('widget-sidecar', 'widget:v2')]
  },
  {
    'name': 'widget',
    'image': 'nginx:latest',
    'init_container_image': [('widget-init', 'widget:v2')],
  },
  {
    'name': 'widget',
    'container_image': [('widget-init', 'widget:v2')],
    'init_container_image': [('widget-init', 'widget:v2')],
  }
]


@pytest.mark.unit
@pytest.mark.parametrize('invalid_kwargs', invalid_kwargs)
def test_invalid_args_raise_errors(invalid_kwargs, mocker):
  mock_client = mocker.MagicMock()

  # DuploError is generic. It doesn't specifically identify bad input. We have to match the error message or we
  # can get false positives if the method raises a DuploError later about something else.
  with pytest.raises(DuploError, match='Provide a service image, container images, or init container images.'):
    DuploService(mock_client).update_image(**invalid_kwargs)


no_matching_container_tests = [
  (
    {
      'Template': {'OtherDockerConfig': json.dumps({'additionalContainers': [
        {
          'name': 'widget-sidecar',
          'image': 'widget:v1',
        }
      ]})}
    },
    {
      'name': 'widget',
      'container_image': [('not-widget-sidecar', 'widget:v2')]
    }
  ),
  (
    {
      'Template': {'OtherDockerConfig': json.dumps({'initContainers': [
        {
          'name': 'widget-init',
          'image': 'widget:v1',
        }
      ]})}
    },
    {
      'name': 'widget',
      'init_container_image': [('not-widget-init', 'widget:v2')]
    }
  )
]


@pytest.mark.unit
@pytest.mark.parametrize('service_definition,kwargs', no_matching_container_tests)
def test_no_matching_container_raises_error(service_definition, kwargs, mocker):
  mocker.patch(
    'duplo_resource.service.DuploService.find',
    mocker.MagicMock(return_value=service_definition)
  )
  mock_client = mocker.MagicMock()

  # DuploError is generic. It doesn't specifically identify bad input. We have to match the error message or we
  # can get false positives if the method raises a DuploError later about something else.
  with pytest.raises(DuploError, match=r'No matching containers found in service .*'):
    DuploService(mock_client).update_image(**kwargs)



post_data_tests = [
  # Update one service.
  (
    {'Template': {}},
    {
      'name': 'widget',
      'image': 'widget:v1'
    },
    {
      'Name': 'widget',
      'Image': 'widget:v1',
      'AllocationTags': '',
    },
  ),

  # Update one additional container.
  (
    {
      'Template': {'OtherDockerConfig': json.dumps({'additionalContainers': [
        {
          'name': 'widget-sidecar',
          'image': 'widget:v1',
        }
      ]})}
    },
    {
      'name': 'widget',
      'container_image': [('widget-sidecar', 'widget:v2')]
    },
    {
      'Name': 'widget',
      'OtherDockerConfig': json.dumps({'additionalContainers': [{
        'name': 'widget-sidecar',
        'image': 'widget:v2',
      }]}),
      'AllocationTags': '',
    },
  ),

  # Update first of two additional containers.
  (
    {
      'Template': {'OtherDockerConfig': json.dumps({'additionalContainers': [
        {
          'name': 'widget-sidecar1',
          'image': 'widget:v1',
        },
        {
          'name': 'widget-sidecar2',
          'image': 'widget:v1',
        }
      ]})}
    },
    {
      'name': 'widget',
      'container_image': [('widget-sidecar1', 'widget:v2')]
    },
    {
      'Name': 'widget',
      'OtherDockerConfig': json.dumps({'additionalContainers': [
        {
          'name': 'widget-sidecar1',
          'image': 'widget:v2',
        },
        {
          'name': 'widget-sidecar2',
          'image': 'widget:v1',
        }
      ]}),
      'AllocationTags': '',
    }
  ),

  # Update second of two additional containers.
  (
    {
      'Template': {'OtherDockerConfig': json.dumps({'additionalContainers': [
        {
          'name': 'widget-sidecar1',
          'image': 'widget:v1',
        },
        {
          'name': 'widget-sidecar2',
          'image': 'widget:v1',
        }
      ]})}
    },
    {
      'name': 'widget',
      'container_image': [('widget-sidecar2', 'widget:v2')]
    },
    {
      'Name': 'widget',
      'OtherDockerConfig': json.dumps({'additionalContainers': [
        {
          'name': 'widget-sidecar1',
          'image': 'widget:v1',
        },
        {
          'name': 'widget-sidecar2',
          'image': 'widget:v2',
        }
      ]}),
      'AllocationTags': '',
    }
  ),

  # Update two of two additional containers.
  (
    {
      'Template': {'OtherDockerConfig': json.dumps({'additionalContainers': [
        {
          'name': 'widget-sidecar1',
          'image': 'widget:v1',
        },
        {
          'name': 'widget-sidecar2',
          'image': 'widget:v1',
        }
      ]})}
    },
    {
      'name': 'widget',
      'container_image': [('widget-sidecar1', 'widget:v2'), ('widget-sidecar2', 'widget:v2')]
    },
    {
      'Name': 'widget',
      'OtherDockerConfig': json.dumps({'additionalContainers': [
        {
          'name': 'widget-sidecar1',
          'image': 'widget:v2',
        },
        {
          'name': 'widget-sidecar2',
          'image': 'widget:v2',
        }
      ]}),
      'AllocationTags': '',
    }
  ),

  # Update two of one additional containers.
  # Arguably, trying to update images for two additional containers in a service that only has one additional container
  # should raise an error. This test was written to assert existing behavior to preserve backwards compatibility.
  (
    {
      'Template': {'OtherDockerConfig': json.dumps({'additionalContainers': [
        {
          'name': 'widget-sidecar2',
          'image': 'widget:v1',
        }
      ]})}
    },
    {
      'name': 'widget',
      'container_image': [('widget-sidecar1', 'widget:v2'), ('widget-sidecar2', 'widget:v2')]
    },
    {
      'Name': 'widget',
      'OtherDockerConfig': json.dumps({'additionalContainers': [
        {
          'name': 'widget-sidecar2',
          'image': 'widget:v2',
        }
      ]}),
      'AllocationTags': '',
    }
  ),

  # Update one init container.
  (
    {
      'Template': {'OtherDockerConfig': json.dumps({'initContainers': [
        {
          'name': 'widget-init',
          'image': 'widget:v1',
        }
      ]})}
    },
    {
      'name': 'widget',
      'init_container_image': [('widget-init', 'widget:v2')]
    },
    {
      'Name': 'widget',
      'OtherDockerConfig': json.dumps({'initContainers': [{
        'name': 'widget-init',
        'image': 'widget:v2',
      }]}),
      'AllocationTags': '',
    },
  ),

  # Update first of two init containers.
  (
    {
      'Template': {'OtherDockerConfig': json.dumps({'initContainers': [
        {
          'name': 'widget-init1',
          'image': 'widget:v1',
        },
        {
          'name': 'widget-init2',
          'image': 'widget:v1',
        }
      ]})}
    },
    {
      'name': 'widget',
      'init_container_image': [('widget-init1', 'widget:v2')]
    },
    {
      'Name': 'widget',
      'OtherDockerConfig': json.dumps({'initContainers': [
        {
          'name': 'widget-init1',
          'image': 'widget:v2',
        },
        {
          'name': 'widget-init2',
          'image': 'widget:v1',
        }
      ]}),
      'AllocationTags': '',
    }
  ),

  # Update second of two init containers.
  (
    {
      'Template': {'OtherDockerConfig': json.dumps({'initContainers': [
        {
          'name': 'widget-init1',
          'image': 'widget:v1',
        },
        {
          'name': 'widget-init2',
          'image': 'widget:v1',
        }
      ]})}
    },
    {
      'name': 'widget',
      'init_container_image': [('widget-init2', 'widget:v2')]
    },
    {
      'Name': 'widget',
      'OtherDockerConfig': json.dumps({'initContainers': [
        {
          'name': 'widget-init1',
          'image': 'widget:v1',
        },
        {
          'name': 'widget-init2',
          'image': 'widget:v2',
        }
      ]}),
      'AllocationTags': '',
    }
  ),

  # Update two of two init containers.
  (
    {
      'Template': {'OtherDockerConfig': json.dumps({'initContainers': [
        {
          'name': 'widget-init1',
          'image': 'widget:v1',
        },
        {
          'name': 'widget-init2',
          'image': 'widget:v1',
        }
      ]})}
    },
    {
      'name': 'widget',
      'init_container_image': [('widget-init1', 'widget:v2'), ('widget-init2', 'widget:v2')]
    },
    {
      'Name': 'widget',
      'OtherDockerConfig': json.dumps({'initContainers': [
        {
          'name': 'widget-init1',
          'image': 'widget:v2',
        },
        {
          'name': 'widget-init2',
          'image': 'widget:v2',
        }
      ]}),
      'AllocationTags': '',
    }
  ),

  # Update two of one init containers.
  # Arguably, trying to update images for two init containers in a service that only has one init container should
  # raise an error. When init containers were added, support for additional containers already existed and it allowed
  # this case. Because images for additional containers and init containers work the same, support for init containers
  # followed the pattern set for additional containers.
  (
    {
      'Template': {'OtherDockerConfig': json.dumps({'initContainers': [
        {
          'name': 'widget-init2',
          'image': 'widget:v1',
        }
      ]})}
    },
    {
      'name': 'widget',
      'init_container_image': [('widget-init1', 'widget:v2'), ('widget-init2', 'widget:v2')]
    },
    {
      'Name': 'widget',
      'OtherDockerConfig': json.dumps({'initContainers': [
        {
          'name': 'widget-init2',
          'image': 'widget:v2',
        }
      ]}),
      'AllocationTags': '',
    }
  )
]


@pytest.mark.unit
@pytest.mark.parametrize('service_definition,kwargs,post_data', post_data_tests)
def test_post_data(service_definition, kwargs, post_data, mocker):
  service = service_definition
  mocker.patch(
    'duplo_resource.service.DuploService.find',
    mocker.MagicMock(return_value=service)
  )
  post_data = post_data
  mock_client = mocker.MagicMock()
  mock_client.wait = False
  DuploService(mock_client).update_image(**kwargs)
  mock_client.post.assert_called_once_with(ANY, post_data)



import os
import sys
import shutil
import logging
import re
from jinja2 import Template
from jinja2.filters import FILTERS
from duplocloud.commander import ep
import duplocloud.args as args
HERE = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.dirname(HERE))
from project import Project, REPO_URL

log = logging.getLogger('mkdocs')
logging.basicConfig(level=logging.INFO)
doc_dir = './dist/docs'
page_meta = None
version = None

resource_nav = []
include_pages = [
  "README.md=index.md",
  "CONTRIBUTING.md",
  "CHANGELOG.md",
  "CODE_OF_CONDUCT.md",
  "SECURITY.md",
  "LICENSE=License.md",
]

ignored = [
  "aws"
]

def copy_static():
  shutil.copytree('./wiki', doc_dir, dirs_exist_ok=True)

def gen_resource_page(endpoint: str):
  cls = endpoint.value.split(':')[-1]
  kind = re.sub(r'^Duplo', '', cls)
  # these two refs are just so slightly different
  ref = endpoint.value.replace(':', '.')
  page = f"{kind}.md"
  resource_nav.append({kind: page})
  fp = f"{doc_dir}/{page}"
  with open(fp, 'w') as f:
    f.write(f"""---
kind: {kind}
---
::: {ref}
""")
    
def gen_include_page(include):
  parts = include.split('=')
  if len(parts) == 2:
    file, page = parts
  else:
    file = parts[0]
    page = file
  fp = f"{doc_dir}/{page}"
  if not os.path.exists(fp):
    with open(fp, 'w') as f:
      f.write(f"--8<-- \"{file}\"")

def page_meta_filter(input):
  """Filter to access page meta data in markdown"""
  t = Template(input)
  return t.render(**page_meta)

def cli_arg_filter(attr):
  return getattr(args, attr.name)

def list_to_csv_filter(input):
  return ', '.join(str(v) for v in input)

def string_or_class_name_filter(input):
  if isinstance(input, str):
    return input
  else:
    return getattr(input, "__name__", str(input))

def on_startup(**kwargs):
  global version
  project = Project()
  version = str(project.latest_tag)
  os.makedirs('dist/docs', exist_ok=True)
  copy_static()
  for e in ep:
    if e.name not in ignored:
      gen_resource_page(e)
  for f in include_pages:
    gen_include_page(f)
  FILTERS['page_meta'] = page_meta_filter
  FILTERS['cli_arg'] = cli_arg_filter
  FILTERS['list_to_csv'] = list_to_csv_filter
  FILTERS['string_or_class_name'] = string_or_class_name_filter

def on_config(config):
  copy_static()
  config["docs_dir"] = "dist/docs"
  config["nav"].insert(3, {"Resources": resource_nav})
  return config

def on_page_markdown(markdown, page, config, **kwargs):
  """Save the page meta data to be used in the page_meta_filter"""
  global page_meta
  page_meta = page.meta
  t = Template(markdown)
  return t.render(
    version=version, 
    repo_url=REPO_URL, 
    **page_meta)




import json
import pytest
from unittest.mock import ANY
from duplo_resource.service import DuploService
from duplocloud.errors import DuploError

@pytest.mark.unit
def test_create_service(mocker):
    mock_client = mocker.MagicMock()
    service = DuploService(mock_client)
    body = {"Name": "test-service", "Image": "nginx:latest"}
    # Mock the find method to return service details
    mock_service_details = {
        "Name": "test-service",
        "Image": "nginx:latest",
        "Replicaset": "rs-123"
    }
    mocker.patch.object(service, 'find', return_value=mock_service_details)
    # Mock the wait method
    mocker.patch.object(service, 'wait')
    # Enable wait flag
    mock_client.wait = True
    service.create(body)
    mock_client.post.assert_called_once_with(ANY, body)

@pytest.mark.unit
def test_delete_service(mocker):
    mock_client = mocker.MagicMock()
    service = DuploService(mock_client)
    service.delete("test-service")
    mock_client.post.assert_called_once_with(ANY, {"Name": "test-service", "State": "delete"})

@pytest.mark.unit
def test_restart_service(mocker):
    mock_client = mocker.MagicMock()
    service = DuploService(mock_client)
    # Mock the find method to return service details
    mock_service_details = {
        "Name": "test-service",
        "Status": "Running"
    }
    mocker.patch.object(service, 'find', return_value=mock_service_details)
    # Mock the wait method
    mocker.patch.object(service, 'wait')
    # Enable wait flag
    mock_client.wait = True
    service.restart("test-service")
    mock_client.post.assert_called_once_with(ANY)

@pytest.mark.unit
def test_stop_service(mocker):
    mock_client = mocker.MagicMock()
    service = DuploService(mock_client)
    # Mock the find method to return service details
    mock_service_details = {
        "Name": "test-service",
        "Status": "Running"
    }
    mocker.patch.object(service, 'find', return_value=mock_service_details)
    # Mock the wait method
    mocker.patch.object(service, 'wait')
    # Enable wait flag
    mock_client.wait = True
    service.stop("test-service")
    mock_client.post.assert_called_once_with(ANY)

@pytest.mark.unit
def test_start_service(mocker):
    mock_client = mocker.MagicMock()
    service = DuploService(mock_client)
    # Mock the find method to return service details
    mock_service_details = {
        "Name": "test-service",
        "Status": "Stopped"
    }
    mocker.patch.object(service, 'find', return_value=mock_service_details)
    # Mock the wait method
    mocker.patch.object(service, 'wait')
    # Enable wait flag
    mock_client.wait = True
    service.start("test-service")
    mock_client.post.assert_called_once_with(ANY)



from duplocloud import args
from duplocloud.client import DuploClient
from duplocloud.errors import DuploError
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.commander import Command, Resource
import json

@Resource("aws_secret")
class DuploAwsSecret(DuploTenantResourceV3):
  """AWS Secrets Manager Secrets resource.

  This resource allows you to create, find, update, and delete AWS Secrets Manager secrets.

  Usage:
    ```sh
    duploctl aws_secret <cmd> [options]
    ```

  Manages [AWS Secrets Manager](https://aws.amazon.com/secrets-manager/) in the background. 
  """
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo, "aws/secret")
  
  def name_from_body(self, body):
    return body.get("Name", None)
  
  @Command()
  #Implement find with opt-in option to display sensitive data. 
  #Note that using find still returns the sensitive data!
  def find(self, 
           name: args.NAME,
           show_sensitive: args.SHOW_SENSITIVE=False) -> dict:
    """Find as AWS Secretmanager secret by name and return its content

    Usage: cli usage
      ```sh
      duploctl aws_secret find <name>
      ```
    
    Args:
      name: The name of the AWS secret to find.
      show_sensitive: Display value of the secretstring field

    Returns: 
      resource: The AWS secret object.
      
    Raises:
      DuploError: If the AWS secret could not be found.
    """
    # if the name has the prefix we good, otherwise add it
    response = None
    try:
      response = self.duplo.get(self.endpoint(name))
    except DuploError as e:
      # if it wasn't found try with the full prefix
      if e.code == 400:
        response = self.duplo.get(self.endpoint(self._prefix_name(name)))
      else:
        raise e
    if not show_sensitive:
      obfuscated_response = response.json()
      sensitive_len = len(response.json()["SecretString"])
      placeholder = "*"
      obfuscated_response["SecretString"] = placeholder * sensitive_len
      return obfuscated_response
    else:
      return response.json()
  
  @Command()
  def create(self, 
             name: args.NAME=None,
             body: args.BODY=None,
             data: args.DATAMAP=None,
             value: args.CONTENT=None,
             dryrun: args.DRYRUN=False) -> dict:
    """Create an AWS Secretmanager Secret

    Using DuploCloud's native support for AWS Secrets Manager, you can create a new secret. This method acts and feels like how the Kubernetes secrets work within this cli. Supports the secrets value as a string or a key/value JSON object where each value is a string. If you give a JSON object with any key that is not a string, the entire value will be simply a string with a JSON value. The examples below mostly include the `--dry-run` so you can see the output. Simply remove that to actually create the secret.

    Usage: cli usage
      ```sh
      duploctl aws_secret create <name> <args>
      ```

    Example: Create a secret from a datamap
      ```sh
      duploctl aws_secret create mysecret --from-literal foo=bar --from-file some-config.json
      ```

    Example: Create a secret with a value
      ```sh
      duploctl aws_secret create mysecret --value foobarbaz
      ```

    Example: Merge a body with new keys
      Notice the the `--file` flag is set to `-` which means it will read a body file from stdin. Since a name is given, the name in the body file will be replaced with the name given in the command.
      ```sh
      cat awssecret.yaml | duploctl aws_secret create mysecret --file - --from-file some-config.json --from-literal icecream=vanilla --dry-run 
      ```
      Here is what the file body within awssecret.yaml looks like
      ```yaml
      --8<-- "src/tests/data/awssecret.yaml"
      ```
      And then the some-config.json file looks like this
      ```json
      --8<-- "src/tests/files/some-config.json"
      ```
      
    Args:
      name: The name of the AWS Secret to create.
      body: The full body of an AWS Secrets Manager secret for DuploCloud.
      data: A map of key-value pairs to be merged into the SecretString field of the AWS Secretmanager secret. Can't be used with the value argument. A datamap is a combination of all of the `--from-literal` and `--from-file` flags.
      value: The value of the AWS Secretmanager secret. This overwrites the existing value! Can't be used with the data argument.
      dryrun: If true, returns the body that would be sent to the API without actually creating the resource.

    Returns: 
      message: Either a succes message is returned or if --dry-run is passed then the body is what is returned. 
      
    Raises:
      DuploError: If the AWS secret already exists.
    """
    # Make sure the user passes name and value, or a body (from file).
    if not name and not body:
      raise DuploError("Name is required when body is not provided")
    if value and data: 
      raise DuploError("You cannot use --value with --from-file or --from-literal. Use one or the other.")
    if not body:
      body = {}
    # If the user passed in a name then this takes precedence over the body name.
    if name:
      body['Name'] = name
    if value:
      body['SecretString'] = value
    elif data:
      # if we have data, then we need to parse the SecretString as json so we can merge the data into it
      secretString = body.get('SecretString', '{}')
      body['SecretString'] = self._merge_data(secretString, data)
    if dryrun:
      return body
    else:
      return super().create(body=body)
    
  @Command()
  def update(self, 
             name: args.NAME=None,
             body: args.BODY=None,
             data: args.DATAMAP=None,
             value: args.CONTENT=None,
             dryrun: args.DRYRUN=False) -> dict:
    """Update an AWS Secretmanager secret.

    Follows all the same arguments and style of the create method. This requires the secret to already exist. 

    Usage: cli usage
      ```sh
      duploctl aws_secret update <name> <args>
      ```
    
    Args:
      name: The name of the AWS Secret to create.
      body: The full body of an AWS Secrets Manager secret for DuploCloud.
      data: A map of key-value pairs to be merged into the SecretString field of the AWS Secrets Manager secret. Cannot be used with the value argument. A datamap is a combination of all of the `--from-literal` and `--from-file` flags.
      value: The value of the AWS Secrets Manager secret. OVERWRITES the existing value. Cannot be used with the data argument.
      dryrun: If true, returns the body that would be sent to the API without actually creating the resource.

    Returns: 
      message: Either a succes message is returned or if --dry-run is passed then the body is what is returned. 
      
    Raises:
      DuploError: If the AWS secret could not be found or doesn't exist.
    """
    if not name and not body:
      raise DuploError("Name is required when body is not provided")
    if value and data: 
      raise DuploError("You cannot pass both value and data. Use one or the other.")
    if not body:
      body = {}
    # If the user passed in a name then this takes precedence over the body name.
    if not name:
      name = self.name_from_body(body)
      if not name:
        raise DuploError("Name is required when the body does not have a name")
    else:
      body['Name'] = name
    # even if we don't use this, it confirms the secret exists before we try and update it
    current = self.find(name, True)
    if value:
      body['SecretString'] = value
    elif data:
      # start with the data from the secret we found
      secretString = current.get('SecretString', '{}')
      body['SecretString'] = self._merge_data(secretString, data)
    body['SecretValueType'] = "plain"
    if dryrun:
      return body
    else:
      return super().update(name=name, body=body)

  @Command()
  def delete(self,
             name: args.NAME) -> dict:
    """Delete an AWS Secretmanager secret.

    Deletes an AWS Secretmanager secret by name.

    Usage: cli
      ```sh
      duploctl aws_secret delete <name>
      ```

    Args:
      name: The name of an AWS Secretmanager secret to delete. This can either be the short name or the full name including the tenant prefix.
      wait: Wait for an AWS Secretmanager secret to be deleted.

    Returns:
      message: A success message.
    """
    # if the name has the prefix we good, otherwise add it
    try:
      super().delete(name)
    except DuploError as e:
      # if it wasn't found try with the full prefix
      if e.code == 404:
        super().delete(self._prefix_name(name))
      else:
        raise e
    return {
      "message": f"Successfully deleted secret '{name}'"
    }
  
  def _merge_data(self, secretString: str, data: dict) -> dict: 
    try:
      secretData = json.loads(secretString)
    except json.JSONDecodeError:
      raise DuploError("Unable to parse the secret string to json. To use --from-file or --from-literal, the SecretString must be a valid JSON object encoded as a string where each key and value is a string.")
    # all of the values in the dict must be a string or we need to throw an error
    if not all(isinstance(v, str) for v in secretData.values()):
      raise DuploError("All values in the existing json object must be strings to use --from-file or --from-literal")
    # now let's merge and set this puppy, woof
    secretData.update(data)
    return json.dumps(secretData)

  # adds the prefix only if the name does not start with a slash
  def _prefix_name(self, name: str) -> str:
    dashed = "" if name.startswith("/") else "-"
    return f"duploservices-{self.duplo.tenant}{dashed}{name}"



from duplocloud.client import DuploClient
from duplocloud.errors import DuploError
from sys import exit

def main():
  try:
    duplo, args = DuploClient.from_env()
    o = duplo(*args)
    if o:
      print(o)
  except DuploError as e:
    print(e)
    exit(e.code)
  except Exception as e:
    print(f"An unexpected error occurred: {e}")
    exit(1)

if __name__ == "__main__":
  main()



import pytest

from duplocloud.errors import DuploError
from duplocloud.client import DuploClient

duplo, _ = DuploClient.from_env()

class Testportal:

  @pytest.mark.integration
  def test_system_info(self):
    r = duplo.load("system")
    try:
      info = r("info")
      assert info["DefaultAwsPartition"] == "aws"
      # Check is EnabledFlags has atleast 1 value.
      assert len(info["EnabledFlags"]) > 0
      # Check is AppConfigs has atleast 1 value.
      assert len(info["AppConfigs"]) > 0
    except DuploError as e:
      pytest.fail(f"Failed to grab system info: {e}")
    # there is at least some information returned.
    assert len(info) > 0




from duplocloud.client import DuploClient  # Importing necessary modules
from duplocloud.errors import DuploFailedResource, DuploStillWaiting
from duplocloud.resource import DuploTenantResourceV3
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("job")  # Decorator to define a resource
class DuploJob(DuploTenantResourceV3):
  """Manage Duplo Kubernetes Jobs

  Duplo Jobs provide a way to run containerized tasks to completion in a Kubernetes cluster.

  See more details at: https://docs.duplocloud.com/docs/kubernetes-overview/jobs
  """
  def __init__(self, duplo: DuploClient):  # Constructor method
    super().__init__(duplo, "k8s/job")  
    self.wait_timeout = 1000
    self.wait_poll = 3
    self.__pod_svc = self.duplo.load("pod")

  @Command()  
  def create(self, 
             body: args.BODY):
    """Create a Kubernetes Job.

    Creates a new Job with the specified configuration. The Job will create pods
    and ensure they complete successfully according to the completion criteria.

    Usage: CLI Usage
      ```sh
      duploctl job create -f job.yaml
      ```
      Contents of the `job.yaml` file
      ```yaml
      --8<-- "src/tests/data/job.yaml"
      ```
    
    Example: Create a Job using a one-liner.
      ```sh
      echo \"\"\"
      --8<-- "src/tests/data/job.yaml"
      \"\"\" | duploctl job create -f -
      ```

    Example: Create a Job using a file.
      ```sh
      duploctl job create -f job.yaml
      ```

    Args:
      body: The complete Job configuration including container specs, completions,
            and other parameters.

    Returns:
      message: Success message confirming the Job creation.

    Raises:
      DuploError: If the Job could not be created due to invalid configuration.
      DuploFailedResource: If the Job's pods encounter faults during execution.
    """
    name = self.name_from_body(body)
    a = 0
    s = 0
    f = 0
    def check_pod_faults(pod, faults):
      fs = [f for f in faults if f["Resource"].get("Name", None) == pod["InstanceId"]]
      fsct = len(fs)
      if fsct > 0:
        d = "\n".join([f["Description"] for f in fs])
        raise DuploFailedResource(f"Pod {pod['InstanceId']} raised {fsct} faults.\n{d}")
    def wait_check():
      nonlocal a, s, f
      job = self.find(name)
      active = job["status"].get("active", 0)
      succeeded = job["status"].get("succeeded", 0)
      failed = job["status"].get("failed", 0)
      completions = job["spec"].get("completions", 1)
      limit = job["spec"].get("backoffLimit", 6)
      cond = job["status"].get("conditions", [])
      cpl = [c for c in cond if c["type"] == "Complete" and c["status"] == "True"]
      fail = [c for c in cond if c["type"] == "Failed" and c["status"] == "True"]
      if (a != active or s != succeeded or f != failed):
        a = active
        s = succeeded
        f = failed
        self.duplo.logger.warn(f"Job {name}: active({active}/{completions}), succeeded({succeeded}/{completions}), failed({failed}/{limit})")
      # make sure we can get pods and logs first
      pods_exist = (active > 0 or succeeded > 0 or failed > 0)
      pods = self.pods(name)
      podct = len(pods)
      if pods_exist and podct == 0:
        raise DuploStillWaiting(f"Job {name} has no pods {pods_exist} {podct}")
      # check for any faults and show the logs if we can
      faults = self.tenant_svc.faults(id=self.tenant_id)
      for pod in pods:
        check_pod_faults(pod, faults)
        self.__pod_svc.logs(pod=pod)
      # make sure we got all of the logs
      pod_count = active + succeeded + failed
      if podct != pod_count:
        raise DuploStillWaiting(f"Expected {pod_count} pods, got {podct}")
      if len(fail) > 0:
        raise DuploFailedResource(f"Job {name} failed with {fail[0]['reason']}: {fail[0]['message']}")
      
      # if none have completed, keep waiting
      if not len(cpl) > 0:
        raise DuploStillWaiting(f"Job '{name}' is waiting for 'Complete' condition")
    super().create(body, wait_check)
    return {
      "message": f"Job {name} ran successfully."
    }


  @Command()
  def pods(self, 
           name: args.NAME):
    """List pods for a Job.

    Retrieve all pods that are managed by the specified Job. The pods are filtered by the Job name.

    Usage: CLI Usage
      ```sh
      duploctl job pods <name>
      ```

    Args:
      name: The name of the Job to get pods for.

    Returns:
      list: A list of pods associated with the Job, including their status and metadata.

    Raises:
      DuploError: If the Job could not be found or if there's an error retrieving pods.
    """
    pods = self.__pod_svc.list()
    return [
      pod for pod in pods
      if pod.get("Name", None) == name and pod["ControlledBy"]["QualifiedType"] == "kubernetes:batch/v1/Job"
    ]



from duplocloud.client import DuploClient
from duplocloud.resource import DuploTenantResourceV2
from duplocloud.errors import DuploError, DuploFailedResource, DuploStillWaiting
from duplocloud.commander import Command, Resource
import duplocloud.args as args

@Resource("hosts")
class DuploHosts(DuploTenantResourceV2):
  """Manage Duplo Hosts
  
  Duplo hosts are virtual machines that run your services within a tenant. You can perform
  lifecycle operations like creating, deleting, starting, stopping and rebooting hosts.
  Hosts can be created with specific configurations like instance type, AMI, and other
  parameters.

  See more details at: https://docs.duplocloud.com/docs/welcome-to-duplocloud/application-focussed-interface/duplocloud-common-components/hosts
  """
  
  def __init__(self, duplo: DuploClient):
    super().__init__(duplo)
    self.paths = {
      "list": "GetNativeHosts"
    }
  
  @Command()
  def create(self,
             body: args.BODY) -> dict:
    """Create a Host resource.

    Creates a new host in the tenant with the specified configuration.

    Usage: CLI Usage
      ```sh
      duploctl hosts create -f 'hosts.yaml'
      ```
      Contents of the `hosts.yaml` file
      ```yaml
      --8<-- "src/tests/data/hosts.yaml"
      ```

    Example: Create a host using a one-liner
      ```sh
      echo \"\"\"
      --8<-- "src/tests/data/hosts.yaml"
      \"\"\" | duploctl hosts create -f -
      ```
    
    Args:
      body: The host configuration including instance type, AMI, and other parameters.
      wait: Whether to wait for the host to be fully created and running.

    Returns: 
      message: Success message and the instance ID of the created host.

    Raises:
      DuploError: If the host could not be created or the configuration is invalid.
      DuploFailedResource: If the host creation process fails.
    """
    def wait_check():
      name = self.name_from_body(body)
      h = self.find(name)
      if h["Status"] != "running":
        if h["Status"] != "pending":
          raise DuploFailedResource(f"Host '{name}' failed to create.")
        raise DuploStillWaiting(f"Host '{name}' is waiting")
    # let's get started
    if body.get("ImageId", None) is None:
      body["ImageId"] = self.discover_image(body.get("AgentPlatform", 0))
    res = self.duplo.post(self.endpoint("CreateNativeHost"), body)
    if self.duplo.wait:
      self.wait(wait_check)
    return {
      "message": f"Successfully created host '{body['FriendlyName']}'",
      "id": res.json()
    }
  
  @Command()
  def delete(self,
             name: args.NAME) -> dict:
    """Delete a host.
    
    Terminates a host by its name. This operation is irreversible and will
    destroy all data on the instance.

    Usage: CLI Usage
      ```sh
      duploctl hosts delete <name>
      ```
    
    Args:
      name: The name of the host to delete.
      wait: Whether to wait for the host to be fully terminated.

    Returns:
      message: Success message confirming the host deletion.

    Raises:
      DuploError: If the host does not exist or cannot be deleted.
      DuploFailedResource: If the deletion process fails.
    """
    host = self.find(name)
    inst_id = host["InstanceId"]
    res = self.duplo.post(self.endpoint(f"TerminateNativeHost/{inst_id}"), host)
    def wait_check():
      h = None 
      try:
        h = self.find(name)
      except DuploError as e:
        if e.code == 404:
          return None # if 404 then it's gone so finish waiting
        else:
          raise DuploFailedResource(f"Host '{name}' failed to delete.")
      if h["Status"] == "shutting-down" or h["Status"] == "running":
        raise DuploStillWaiting(f"Host '{name}' is waiting for termination")
    if self.duplo.wait:
      self.wait(wait_check, 500)
    return {
      "message": f"Successfully deleted host '{name}'",
      "data": res.json()
    }
  
  @Command()
  def stop(self,
           name: args.NAME) -> dict:
    """Stop a host.
    
    Stops a running host. The instance can be restarted later using the start
    command. Stopped instances do not incur compute charges but still incur storage costs.

    Usage: CLI Usage
      ```sh
      duploctl hosts stop <name>
      ```
    
    Args:
      name: The name of the host to stop.
      wait: Whether to wait for the host to reach stopped state.

    Returns:
      message: Success message confirming the host has been stopped.

    Raises:
      DuploError: If the host does not exist or cannot be stopped.
      DuploFailedResource: If the stop process fails.
    """
    host = self.find(name)
    inst_id = host["InstanceId"]
    res = self.duplo.post(self.endpoint(f"stopNativeHost/{inst_id}"), host)
    def wait_check():
      h = self.find(name)
      if h["Status"] == "running":
        raise DuploError(f"Host '{name}' not ready", 404)
      if h["Status"] != "stopped":
        if h["Status"] != "stopping":
          raise DuploFailedResource(f"Host '{name}' failed to stop.")
        raise DuploStillWaiting(f"Host '{name}' is waiting for status stopped")
    if self.duplo.wait:
      self.wait(wait_check, 500)
    return {
      "message": f"Successfully stopped host '{name}'",
      "data": res.json()
    }
  
  @Command()
  def start(self,
             name: args.NAME) -> dict:
    """Start a host.
    
    Starts a stopped host. The instance will retain its configuration and data
    from when it was stopped.

    Usage: CLI Usage
      ```sh
      duploctl hosts start <name>
      ```
    
    Args:
      name: The name of the host to start.
      wait: Whether to wait for the host to reach running state.
    
    Returns:
      message: Success message confirming the host has been started.

    Raises:
      DuploError: If the host does not exist or cannot be started.
      DuploFailedResource: If the start process fails.
    """
    host = self.find(name)
    inst_id = host["InstanceId"]
    res = self.duplo.post(self.endpoint(f"startNativeHost/{inst_id}"), host)
    def wait_check():
      h = self.find(name)
      if h["Status"] == "stopped":
        raise DuploError(f"Host '{name}' not ready", 404)
      if h["Status"] != "running":
        if h["Status"] != "pending":
          raise DuploFailedResource(f"Host '{name}' failed to stop.")
        raise DuploStillWaiting(f"Host '{name}' is waiting for status running")
    if self.duplo.wait:
      self.wait(wait_check, 500)
    return {
      "message": f"Successfully started host '{name}'",
      "data": res.json()
    }
  
  @Command()
  def reboot(self,
             name: args.NAME) -> dict:
    """Reboot a host.
    
    Performs a graceful reboot of a host. This is equivalent to an operating
    system reboot command. The instance ID and data are preserved.

    Usage: CLI Usage
      ```sh
      duploctl hosts reboot <name>
      ```

    Args:
      name: The name of the host to reboot.
    
    Returns:
      message: Success message confirming the reboot request.

    Raises:
      DuploError: If the host does not exist or cannot be rebooted.
    """
    host = self.find(name)
    inst_id = host["InstanceId"]
    res = self.duplo.post(self.endpoint(f"RebootNativeHost/{inst_id}"), host)
    return {
      "message": f"Successfully rebooted host '{name}'",
      "data": res.json()
    }
    
  def name_from_body(self, body):
    prefix = f"duploservices-{self.tenant['AccountName']}"
    name =  body["FriendlyName"]
    if not name.startswith(prefix):
      name = f"{prefix}-{name}"
    return name
  
  def discover_image(self, agent, arch="amd64"):
    imgs = self.tenant_svc.host_images(self.tenant['AccountName'])
    try:
      img = [i for i in imgs if i["Agent"] == agent and i["Arch"] == arch][0]
      return img.get("ImageId")
    except IndexError:
      raise DuploError(f"Image for agent '{agent}' not found", 404)



import pytest
from duplocloud.errors import DuploError, DuploExpiredCache, DuploFailedResource, DuploStillWaiting

@pytest.mark.unit
def test_duplo_error_defaults():
    err = DuploError("Something went wrong")
    assert isinstance(err, DuploError)
    assert str(err) == "Something went wrong"
    assert err.code == 1
    assert err.response is None

@pytest.mark.unit
def test_duplo_error_custom_code():
    err = DuploError("Custom code", code=500)
    assert err.code == 500

@pytest.mark.unit
def test_duplo_expired_cache():
    key = "test_key"
    err = DuploExpiredCache(key)
    assert isinstance(err, DuploExpiredCache)
    assert isinstance(err, DuploError)
    assert err.key == key
    assert err.code == 404
    assert str(err) == "Cache item {key} is expired"

@pytest.mark.unit
def test_duplo_failed_resource():
    err = DuploFailedResource("test-resource")
    assert isinstance(err, DuploFailedResource)
    assert err.code == 412
    assert str(err) == "test-resource is in a failed state"

@pytest.mark.unit
def test_duplo_still_waiting():
    err = DuploStillWaiting("delayed-service")
    assert isinstance(err, DuploStillWaiting)
    assert err.code == 408
    assert str(err) == "delayed-service"

@pytest.mark.integration
@pytest.mark.parametrize("exception_class,init_arg,expected_str,expected_code", [
    (DuploExpiredCache, "mykey", "Cache item {key} is expired", 404),
    (DuploFailedResource, "resource-A", "resource-A is in a failed state", 412),
    (DuploStillWaiting, "resource-B", "resource-B is in a waiting state", 408),
])
def test_integration_duplo_exceptions(exception_class, init_arg, expected_str, expected_code):
    err = exception_class(init_arg)
    assert isinstance(err, DuploError)
    assert str(err) == expected_str
    assert err.code == expected_code
